# SPDX-FileCopyrightText: 2021 Kevin Matocha, Tim C, Jose David M
#
# SPDX-License-Identifier: MIT

"""
`adafruit_displayio_layout.widgets`
=======================
"""

import vectorio

try:
    import bitmaptools
except ImportError:
    pass


# pylint: disable=invalid-name, too-many-arguments
def rectangle_helper(
    x0: int,
    y0: int,
    height: int,
    width: int,
    bitmap,
    color_index: int,
    palette,
    bitmaptool: bool = True,
) -> None:
    """rectangle_helper function
    Draws a rectangle to the bitmap given using ``bitmapstools.bitmap`` or
    ``vectorio.rectangle`` functions

    :param int x0: rectangle lower corner x position
    :param int y0: rectangle lower corner y position

    :param int width: rectangle upper corner x position
    :param int height: rectangle upper corner y position

    :param int color_index: palette color index to be used
    :param palette: palette object to be used to draw the rectangle

    :param bitmap: bitmap for the rectangle to be drawn
    :param bool bitmaptool: uses :py:func:`~bitmaptools.draw_line` to draw the rectanlge.
     when `False` uses :py:func:`~vectorio.Rectangle`

    :return: None
    :rtype: None

             ┌───────────────────────┐
             │                       │
             │                       │
     (x0,y0) └───────────────────────┘

    """
    if bitmaptool:
        bitmaptools.fill_region(bitmap, x0, y0, x0 + width, y0 + height, color_index)
    else:
        rect = vectorio.Rectangle(width, height)
        vectorio.VectorShape(
            shape=rect,
            pixel_shader=palette,
            x=x0,
            y=y0,
        )



# SPDX-FileCopyrightText: 2021 Jose David M.
#
# SPDX-License-Identifier: MIT
#############################
"""
This is a basic demonstration of a Cartesian widget.
"""

import time
import board
import displayio
import terminalio
from adafruit_displayio_layout.widgets.cartesian import Cartesian

# Fonts used for the Dial tick labels
tick_font = terminalio.FONT

display = board.DISPLAY  # create the display on the PyPortal or Clue (for example)
# otherwise change this to setup the display
# for display chip driver and pinout you have (e.g. ILI9341)


# Create a Cartesian widget
my_plane = Cartesian(
    x=150,  # x position for the plane
    y=100,  # y plane position
    width=100,  # display width
    height=100,  # display height
    axes_color=0xFFFFFF,  # axes line color
    axes_stroke=2,  # axes lines width in pixels
    tick_color=0xFFFFFF,  # ticks color
    major_tick_stroke=1,  # ticks width in pixels
    major_tick_length=5,  # ticks length in pixels
    tick_label_font=tick_font,  # the font used for the tick labels
    font_color=0xFFFFFF,  # ticks line color
)

my_group = displayio.Group()
my_group.append(my_plane)
display.show(my_group)  # add high level Group to the display

posx = 0
posy = 0

while True:
    for i in range(0, 90, 2):
        my_plane.update_pointer(i, i)
        time.sleep(0.5)



# SPDX-FileCopyrightText: 2021 Jose David M.
#
# SPDX-License-Identifier: MIT
#############################
"""
This is a more advance demonstration of a Cartesian widget and some configurable
parameters.
"""

import board
import displayio
import terminalio
from adafruit_displayio_layout.widgets.cartesian import Cartesian

# Fonts used for the Dial tick labels
tick_font = terminalio.FONT

display = board.DISPLAY  # create the display on the PyPortal or Clue (for example)
# otherwise change this to setup the display
# for display chip driver and pinout you have (e.g. ILI9341)


# Create different Cartesian widgets
my_group = displayio.Group()

car = Cartesian(
    x=25,
    y=10,
    width=100,
    height=100,
    subticks=True,
)
my_group.append(car)

car3 = Cartesian(
    x=150,
    y=10,
    width=150,
    height=100,
    xrange=(0, 160),
    axes_stroke=1,
    axes_color=0x990099,
    subticks=True,
)
my_group.append(car3)

car4 = Cartesian(
    x=30,
    y=140,
    width=80,
    height=80,
    axes_stroke=1,
    tick_color=0xFFFFFF,
    subticks=True,
)

my_group.append(car4)

car5 = Cartesian(
    x=180,
    y=140,
    width=70,
    height=70,
    xrange=(0, 120),
    yrange=(0, 90),
    tick_color=0x990099,
    axes_stroke=3,
    major_tick_length=10,
)
my_group.append(car5)

display.show(my_group)

while True:
    pass



# SPDX-FileCopyrightText: 2021 Kevin Matocha
#
# SPDX-License-Identifier: MIT
"""

`icon_animated`
================================================================================
A touch enabled widget that includes an animated icon image with a small text label
centered below it.

* Author(s): Kevin Matocha

Implementation Notes
--------------------

**Hardware:**

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

"""
import gc
import time
from math import pi
import bitmaptools
from displayio import TileGrid, Bitmap, Palette
import adafruit_imageload
from adafruit_displayio_layout.widgets.icon_widget import IconWidget
from adafruit_displayio_layout.widgets.easing import quadratic_easeout as easein
from adafruit_displayio_layout.widgets.easing import quadratic_easein as easeout


class IconAnimated(IconWidget):

    """
    An animated touch enabled widget that holds an icon image loaded with
    OnDiskBitmap and a text label centered beneath it. Includes optional
    animation to increase the icon size when pressed.

    .. Warning::  The `init_class` class function must be called before instancing any
        IconAnimated widgets.

    :param str label_text: the text that will be shown beneath the icon image.
    :param str icon: the filepath of the bmp image to be used as the icon.
    :param bool on_disk: if True use OnDiskBitmap instead of imageload to load static
     icon image. This can be helpful to save memory. (default: False)  Note: Bitmap
     file must use indexed colors to allow animations in the IconAnimated widget.

    :param float scale: the maximum zoom during animation, set 1.0 for no zoom.
     A value of 1.5 is a good starting point. The ``scale`` can be less than
     1.0 for shrink animations. (default: same as ``max_scale`` set in ``init_class``),

    :param float angle: the maximum degrees of rotation during animation, positive values
     are clockwise, set 0 for no rotation, in degrees (default: 4 degrees)
    :param float animation_time: the time for the animation in seconds, set to 0.0 for
     no animation, a value of 0.15 is a good starting point (default: 0.15 seconds)

    :param int x: x location the icon widget should be placed. Pixel coordinates.
    :param int y: y location the icon widget should be placed. Pixel coordinates.
    :param anchor_point: (X,Y) values from 0.0 to 1.0 to define the anchor point relative to the
     widget bounding box
    :type anchor_point: Tuple[float,float]
    :param int anchored_position: (x,y) pixel value for the location of the anchor_point
    :type anchored_position: Tuple[int, int]
    """

    # pylint: disable=bad-super-call, too-many-instance-attributes, too-many-locals
    # pylint: disable=too-many-arguments, unused-argument

    display = None
    # The other Class variables are created in Class method `init_class`:
    #               max_scale, bitmap_buffer, palette_buffer

    @classmethod
    def init_class(
        cls, display=None, max_scale=1.5, max_icon_size=(80, 80), max_color_depth=256
    ):
        """
        Initializes the IconAnimated Class variables, including preallocating memory
        buffers for the icon zoom bitmap and icon zoom palette.

        .. Note::  The `init_class` class function must be called before instancing any
                IconAnimated widgets. Usage example:
                ``IconAnimated.init_class(display=board.DISPLAY, max_scale=1.5,
                max_icon_size=(80,80), max_color_depth=256)``

        :param displayio.Display display: The display where the icons will be displayed.
        :param float max_scale: The maximum zoom of the any of the icons, should be >= 1.0,
         (default: 1.5)
        :param max_icon_size: The maximum (x,y) pixel dimensions of any `IconAnimated` bitmap size
         that will be created (default: (80,80)).  Note: This is the original pixel size,
         before scaling
        :type max_icon_size: Tuple[int,int]
        :param int max_color_depth: The maximum color depth of any `IconAnimated`
         bitmap that will be created (default: 256)
        """
        if display is None:
            raise ValueError(
                "IconAninmated.init_class: Must provide display parameter for IconAnimated."
            )

        if (
            isinstance(max_icon_size, tuple)
            and len(max_icon_size) == 2  # validate max_icon_size input
            and isinstance(max_icon_size[0], int)
            and isinstance(max_icon_size[1], int)
        ):
            pass
        else:
            raise ValueError(
                "IconAninmated.init_class: max_icon_size must be an (x,y) "
                "tuple of integer pixel sizes."
            )

        cls.display = display
        if max_scale < 1.0:
            print(
                "Warning: IconAnimated.init_class - max_scale value was "
                "constrained to minimum of 1.0"
            )
        cls.max_scale = max(1.0, max_scale)
        cls.bitmap_buffer = Bitmap(
            round(cls.max_scale * max_icon_size[0]),
            round(cls.max_scale * max_icon_size[1]),
            max_color_depth + 1,
        )
        cls.palette_buffer = Palette(max_color_depth + 1)

    def __init__(
        self,
        label_text,
        icon,
        on_disk=False,
        scale=None,
        angle=4,
        animation_time=0.15,
        **kwargs,
    ):

        if self.__class__.display is None:
            raise ValueError(
                "Must initialize class using\n"
                "`IconAnimated.init_class(display, max_scale, max_color_depth)`\n"
                "prior to instancing IconAnimated widgets."
            )

        super().__init__(label_text, icon, on_disk, **kwargs)  # initialize superclasses

        # constrain instance's maximum_scaling between 1.0 and the Class's max_scale
        if scale is None:
            self._scale = self.__class__.max_scale
        else:
            if scale > self.__class__.max_scale:
                print(
                    "Warning - IconAnimated: max_scale is constrained by value of "
                    "IconAnimated.max_scale set by IconAnimated.init_class(): {}".format(
                        self.__class__.max_scale
                    )
                )
            self._scale = max(0, min(scale, self.__class__.max_scale))

        self._animation_time = animation_time  # in seconds
        self._angle = (angle / 360) * 2 * pi  # in degrees, convert to radians
        self._zoomed = False  # state variable for zoom status

    def zoom_animation(self, touch_point):
        """Performs zoom animation when icon is pressed.

        :param touch_point: x,y location of the screen.
        :type touch_point: Tuple[x,y]
        :return: None
        """

        if self._animation_time > 0:
            try:
                _image, _palette = adafruit_imageload.load(self._icon)

                if len(self.__class__.palette_buffer) < len(_palette) + 1:
                    self._animation_time = 0  # skip any animation
                    print(
                        "Warning: IconAnimated - icon bitmap exceeds IconAnimated.max_color_depth;"
                        " defaulting to no animation"
                    )

            except NotImplementedError:
                self._animation_time = 0  # skip any animation
                print(
                    "Warning: IconAnimated - True color BMP unsupported for animation;"
                    " defaulting to no animation"
                )

        if self._animation_time > 0:

            animation_bitmap = self.__class__.bitmap_buffer
            animation_palette = self.__class__.palette_buffer

            # store the current display refresh setting
            refresh_status = self.__class__.display.auto_refresh

            ###
            ## Update the zoom palette and bitmap buffers and append the tilegrid
            ###

            # copy the image palette, add a transparent color at the end
            for i, color in enumerate(_palette):
                animation_palette[i] = color
            animation_palette[len(animation_palette) - 1] = 0x000000
            animation_palette.make_transparent(len(animation_palette) - 1)

            # create the zoom bitmap larger than the original image to allow for zooming
            animation_bitmap.fill(len(animation_palette) - 1)  # transparent fill
            animation_bitmap.blit(
                (animation_bitmap.width - _image.width) // 2,
                (animation_bitmap.height - _image.height) // 2,
                _image,
            )  # blit the image into the center of the zoom_bitmap

            # place zoom_bitmap at same location as image
            animation_tilegrid = TileGrid(
                animation_bitmap, pixel_shader=animation_palette
            )
            animation_tilegrid.x = -(animation_bitmap.width - _image.width) // 2
            animation_tilegrid.y = -(animation_bitmap.height - _image.height) // 2

            self.__class__.display.auto_refresh = False  # set auto_refresh off
            self[0].hidden = True  # hide the original icon
            self.append(animation_tilegrid)  # add to the self group.

            # Animation: zoom larger
            start_time = time.monotonic()

            while True:
                elapsed_time = time.monotonic() - start_time
                position = min(
                    1.0, easein(elapsed_time / self._animation_time)
                )  # fractional position
                animation_bitmap.fill(len(animation_palette) - 1)
                bitmaptools.rotozoom(
                    dest_bitmap=animation_bitmap,
                    ox=animation_bitmap.width // 2,
                    oy=animation_bitmap.height // 2,
                    source_bitmap=_image,
                    px=_image.width // 2,
                    py=_image.height // 2,
                    scale=1.0 + position * (self._scale - 1.0),  # start scaling at 1.0
                    angle=position * self._angle,
                )
                self.__class__.display.refresh()
                if elapsed_time > self._animation_time:
                    break

            # set display.auto_refresh back to original value
            self.__class__.display.auto_refresh = refresh_status

            del _image
            del _palette
            gc.collect()

            self._zoomed = True

    def zoom_out_animation(self, touch_point):
        """Performs un-zoom animation when icon is released.

        :param touch_point: x,y location of the screen.
        :type touch_point: Tuple[x,y]
        :return: None
        """

        if (self._animation_time > 0) and self._zoomed:
            _image, _palette = adafruit_imageload.load(self._icon)
            animation_bitmap = self.__class__.bitmap_buffer
            animation_palette = self.__class__.palette_buffer

            # store the current display refresh setting
            refresh_status = self.__class__.display.auto_refresh

            self.__class__.display.auto_refresh = False  # set auto_refresh off

            # Animation: shrink down to the original size
            start_time = time.monotonic()
            while True:
                elapsed_time = time.monotonic() - start_time
                position = max(0.0, easeout(1 - (elapsed_time / self._animation_time)))
                animation_bitmap.fill(len(animation_palette) - 1)
                bitmaptools.rotozoom(
                    dest_bitmap=animation_bitmap,
                    ox=animation_bitmap.width // 2,
                    oy=animation_bitmap.height // 2,
                    source_bitmap=_image,
                    px=_image.width // 2,
                    py=_image.height // 2,
                    scale=1.0 + position * (self._scale - 1.0),
                    angle=position * self._angle,
                )
                self.__class__.display.refresh()
                if elapsed_time > self._animation_time:
                    break

            # clean up the zoom display elements
            self[0].hidden = False  # unhide the original icon
            self.pop(-1)  # remove zoom tilegrid from the group
            self.__class__.display.refresh()

            # set display.auto_refresh back to original value
            self.__class__.display.auto_refresh = refresh_status

            del _image
            del _palette
            gc.collect()

        self._zoomed = False



# SPDX-FileCopyrightText: 2021 Kevin Matocha
#
# SPDX-License-Identifier: MIT
"""

`flip_input`
================================================================================
A flip style input selector.

* Author(s): Kevin Matocha

Implementation Notes
--------------------

**Hardware:**

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

"""

import gc
import time
import displayio
from terminalio import FONT

from adafruit_display_shapes.triangle import Triangle

from adafruit_display_text import bitmap_label
from adafruit_displayio_layout.widgets.widget import Widget
from adafruit_displayio_layout.widgets.control import Control

# pylint: disable=reimported

# select the two "easing" functions to use for animations
from adafruit_displayio_layout.widgets.easing import back_easeinout as easein
from adafruit_displayio_layout.widgets.easing import back_easeinout as easeout

# pylint: disable=too-many-arguments, too-many-branches, too-many-statements
# pylint: disable=too-many-locals, too-many-instance-attributes


class FlipInput(Widget, Control):
    """A flip style input selector. The value changes based on touch inputs on the
    two halves of the indicator with optional arrows added.


    :param int x: pixel position
    :param int y: pixel position

    :param displayio.Display display: the display where the widget will be displayed
    :param value_list: the list of strings that will be displayed
    :type value_list: List[str]
    :param Font font: the font used for the text (defaults to ``terminalio.FONT``)
    :param int font_scale: the scaling of the font in integer values (default is 1)
    :param int color: the color used for the font (default is 0xFFFFFF)
    :param int value: the index into the value_list that is initially displayed
     (default is 0)

    :param int arrow_color: the color used for the arrow fill (default is 0x333333)
    :param int arrow_outline: the color used for the arrow outline (default is 0x555555)
    :param int arrow_height: the height of the arrows, in pixels (default is 30 pixels)
    :param int arrow_width: the width of the arrows, in pixels (default is 30 pixels)
    :param int arrow_gap: distance from text to the arrow, in pixels (default is 5),
     can also be a negative value
    :param int arrow_touch_padding: additional padding on the arrow sides of the
     widget where touch response is accepted, in pixels (default = 0)
    :param int alt_touch_padding: additional padding on the non-arrow sides of the
     widget where touch response is accepted, in pixels (default = 0)
    :param Boolean horizontal: set to `True` to display arrows are in the horizontal
     direction, set `False` for arrows in the vertical direction (default = `True`)
    :param float animation_time: duration for the animation during flipping between
     values, in seconds (default is 0.4 seconds), set to 0.0 or `None` for no animation.
    :param float cool_down: minimum duration between activations of the widget with a
     continuous pressing, this can be used to reduce the chance of accidental multiple
     activations, in seconds (default is 0.0 seconds, no delay).  Set to -1.0 to require
     the button be released and pressed again for activation (Note: This requires calling
     the ``released`` function prior to the next call to ``selected``.)

    """

    def __init__(
        self,
        display,
        *,
        value_list=None,
        font=FONT,
        font_scale=1,
        color=0xFFFFFF,
        value=0,  # initial value, index into the value_list
        arrow_touch_padding=0,  # additional touch padding on the arrow sides of the Widget
        arrow_color=0x333333,
        arrow_outline=0x555555,
        arrow_height=30,
        arrow_width=30,
        arrow_gap=5,
        alt_touch_padding=0,  # touch padding on the non-arrow sides of the Widget
        horizontal=True,
        animation_time=None,
        cool_down=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # Group elements for the FlipInput.
        # 0. The text
        # 1. The group holding the temporary scroll bitmap
        # 2. Up arrow: Triangle
        # 3. Down arrow: Triangle

        # initialize the Control superclass

        # pylint: disable=bad-super-call
        super(Control, self).__init__()

        self.value_list = value_list
        self._value = value

        self._color = color
        self._font = font
        self._font_scale = font_scale
        # preload the glyphs

        self._arrow_touch_padding = arrow_touch_padding
        self._alt_touch_padding = alt_touch_padding

        self._horizontal = horizontal
        self._display = display

        self._animation_time = animation_time
        self._cool_down = cool_down
        self._last_pressed = time.monotonic()
        self._pressed = False  # state variable

        # Find the maximum bounding box of the text and determine the
        # baseline (x,y) start point (top, left)

        left = None
        right = None
        top = None
        bottom = None

        for this_value in value_list:
            xposition = 0

            for i, character in enumerate(this_value):
                glyph = self._font.get_glyph(ord(character))

                if (
                    i == 0
                ):  # if it's the first character in the string, check the left value
                    if left is None:
                        left = glyph.dx
                    else:
                        left = min(left, glyph.dx)

                if right is None:
                    right = max(
                        xposition + glyph.dx + glyph.width, xposition + glyph.shift_x
                    )
                else:
                    right = max(
                        right,
                        xposition + glyph.dx + glyph.width,
                        xposition + glyph.shift_x,
                    )  # match bitmap_label

                if top is None:
                    top = -(glyph.height + glyph.dy)
                else:
                    top = min(top, -(glyph.height + glyph.dy))

                if bottom is None:
                    bottom = -glyph.dy
                else:
                    bottom = max(bottom, -glyph.dy)

                xposition = xposition + glyph.shift_x

        self._bounding_box = [
            0,
            0,
            (right - left) * self._font_scale,
            (bottom - top) * self._font_scale,
        ]

        # Create the text label
        self._label = bitmap_label.Label(
            text=value_list[value],
            font=self._font,
            scale=self._font_scale,
            color=self._color,
            base_alignment=True,
            background_tight=True,
        )
        self._label.x = -1 * left * self._font_scale
        self._label.y = -1 * top * self._font_scale

        self._left = left
        self._top = top

        self.append(self._label)  # add the label to the self Group

        # set the touch_boundary including the touch_padding
        self._arrow_gap = arrow_gap  # of pixel gap above/below label before the arrow

        if horizontal:  # horizontal orientation, add arrow padding to x-dimension and
            # alt_padding to y-dimension
            self.touch_boundary = [
                self._bounding_box[0]
                - self._arrow_gap
                - arrow_height
                - self._arrow_touch_padding,
                self._bounding_box[1] - self._alt_touch_padding,
                self._bounding_box[2]
                + 2 * (self._arrow_gap + arrow_height + self._arrow_touch_padding),
                self._bounding_box[3] + 2 * self._alt_touch_padding,
            ]
        else:  # vertical orientation, add arrow padding to y-dimension and
            # alt_padding to x-dimension
            self.touch_boundary = [
                self._bounding_box[0] - self._alt_touch_padding,
                self._bounding_box[1]
                - self._arrow_gap
                - arrow_height
                - self._arrow_touch_padding,
                self._bounding_box[2] + 2 * self._alt_touch_padding,
                self._bounding_box[3]
                + 2 * (self._arrow_gap + arrow_height + self._arrow_touch_padding),
            ]

        # create the Up/Down arrows
        self._update_position()  # call Widget superclass function to reposition

        self._animation_group = displayio.Group(
            scale=self._font_scale
        )  # holds the animation bitmap
        # self._animation_group.x = -1 * left * (1)
        # self._animation_group.y = -1 * top * (1)
        self._animation_group.hidden = True
        self.append(self._animation_group)

        # Add the two arrow triangles, if required

        if (arrow_color is not None) or (arrow_outline is not None):

            if horizontal:  # horizontal orientation, add left and right arrows

                if (
                    (arrow_width is not None)
                    and (arrow_height is not None)
                    and (arrow_width > 0)
                ):
                    mid_point_y = self._bounding_box[1] + self._bounding_box[3] // 2
                    self.append(
                        Triangle(
                            self._bounding_box[0] - self._arrow_gap,
                            mid_point_y - arrow_height // 2,
                            self._bounding_box[0] - self._arrow_gap,
                            mid_point_y + arrow_height // 2,
                            self._bounding_box[0] - self._arrow_gap - arrow_width,
                            mid_point_y,
                            fill=arrow_color,
                            outline=arrow_outline,
                        )
                    )

                    self.append(
                        Triangle(
                            self._bounding_box[0]
                            + self._bounding_box[2]
                            + self._arrow_gap,
                            mid_point_y - arrow_height // 2,
                            self._bounding_box[0]
                            + self._bounding_box[2]
                            + self._arrow_gap,
                            mid_point_y + arrow_height // 2,
                            self._bounding_box[0]
                            + self._bounding_box[2]
                            + self._arrow_gap
                            + arrow_width,
                            mid_point_y,
                            fill=arrow_color,
                            outline=arrow_outline,
                        )
                    )
            else:  # vertical orientation, add upper and lower arrows

                if (
                    (arrow_height is not None)
                    and (arrow_width is not None)
                    and (arrow_height > 0)
                ):
                    mid_point_x = self._bounding_box[0] + self._bounding_box[2] // 2
                    self.append(
                        Triangle(
                            mid_point_x - arrow_width // 2,
                            self._bounding_box[1] - self._arrow_gap,
                            mid_point_x + arrow_width // 2,
                            self._bounding_box[1] - self._arrow_gap,
                            mid_point_x,
                            self._bounding_box[1] - self._arrow_gap - arrow_height,
                            fill=arrow_color,
                            outline=arrow_outline,
                        )
                    )
                    self.append(
                        Triangle(
                            mid_point_x - arrow_width // 2,
                            self._bounding_box[1]
                            + self._bounding_box[3]
                            + self._arrow_gap,
                            mid_point_x + arrow_width // 2,
                            self._bounding_box[1]
                            + self._bounding_box[3]
                            + self._arrow_gap,
                            mid_point_x,
                            self._bounding_box[1]
                            + self._bounding_box[3]
                            + self._arrow_gap
                            + arrow_height,
                            fill=arrow_color,
                            outline=arrow_outline,
                        )
                    )

    # Draw function to update the current value
    def _update_value(self, new_value, animate=True):

        if (
            (self._animation_time is not None)
            and (self._animation_time > 0)  # If animation is required
            and (animate)
        ):

            if ((new_value - self.value) == 1) or (
                (self.value == (len(self.value_list) - 1)) and (new_value == 0)
            ):  # wrap around
                start_position = 0.0
                end_position = 1.0
            else:
                start_position = 1.0
                end_position = 0.0

            self._display.auto_refresh = False

            gc.collect()

            # create the animation bitmap
            animation_bitmap = displayio.Bitmap(
                self._bounding_box[2] // self._font_scale,
                self._bounding_box[3] // self._font_scale,
                2,
            )  # color depth 2

            palette = displayio.Palette(2)
            palette.make_transparent(0)
            palette[1] = self._color
            animation_tilegrid = displayio.TileGrid(
                animation_bitmap, pixel_shader=palette
            )

            # add bitmap to the animation_group
            self._animation_group.append(animation_tilegrid)

            # store away the initial starting bitmap
            start_bitmap = displayio.Bitmap(
                self._label.bitmap.width, self._label.bitmap.height, 2
            )  # color depth 2
            start_bitmap.blit(0, 0, self._label.bitmap)

            # get the bitmap1 position offsets
            bitmap1_offset = [
                -1 * self._left + self._label.tilegrid.x,
                -1 * self._top + self._label.tilegrid.y,
            ]

            # hide the label group
            self.pop(0)

            # update the value label and get the bitmap offsets
            self._label.text = str(self.value_list[new_value])
            bitmap2_offset = [
                -1 * self._left + self._label.tilegrid.x,
                -1 * self._top + self._label.tilegrid.y,
            ]

            # animate between old and new bitmaps
            _animate_bitmap(
                display=self._display,
                target_bitmap=animation_bitmap,
                bitmap1=start_bitmap,
                bitmap1_offset=bitmap1_offset,
                bitmap2=self._label.bitmap,
                bitmap2_offset=bitmap2_offset,
                start_position=start_position,
                end_position=end_position,
                animation_time=self._animation_time,
                horizontal=self._horizontal,
            )

            # unhide the label group
            self.insert(0, self._label)

            # hide the animation group
            self._animation_group.pop()
            # free up memory
            del animation_bitmap
            del start_bitmap
            gc.collect()

            # ensure the display will auto_refresh (likely redundant)
            self._display.auto_refresh = True

        else:  # Update with no animation
            self._display.auto_refresh = False
            self._label.text = str(self.value_list[new_value])
            self._display.auto_refresh = True
        self._update_position()  # call Widget superclass function to reposition

    def _ok_to_change(self):  # checks state variable and timers to determine
        # if an update is allowed
        if self._cool_down < 0:  # if cool_down is negative, require ``released``
            # to be called before next change
            return not self._pressed
        if (time.monotonic() - self._last_pressed) < self._cool_down:
            return False  # cool_down time has not transpired
        return True

    def contains(self, touch_point):  # overrides, then calls Control.contains(x,y)
        """Returns True if the touch_point is within the widget's touch_boundary."""

        ######
        #
        # IMPORTANT: The touch_point should be adjusted to local coordinates, by
        # offsetting for self.x and self.y before calling the Control superclass function
        #
        ######
        touch_x = (
            touch_point[0] - self.x
        )  # adjust touch position for the local position
        touch_y = touch_point[1] - self.y

        return super().contains((touch_x, touch_y, 0))

    def selected(self, touch_point):
        """Response function when the Control is selected.  Increases value when upper half
        is pressed and decreases value when lower half is pressed."""

        # Adjust for local position of the widget using self.x and self.y

        if self._ok_to_change():
            t_b = self.touch_boundary

            if self._horizontal:
                if (
                    t_b[0] <= (touch_point[0] - self.x) < (t_b[0] + t_b[2] // 2)
                ):  # in left half of touch_boundary
                    self.value = self.value - 1

                elif (
                    (t_b[0] + t_b[2] // 2)
                    <= (touch_point[0] - self.x)
                    <= (t_b[0] + t_b[2])
                ):  # in right half of touch_boundary
                    self.value = self.value + 1

            else:
                if (
                    t_b[1] <= (touch_point[1] - self.y) < (t_b[1] + t_b[3] // 2)
                ):  # in upper half of touch_boundary
                    self.value = self.value + 1

                elif (
                    (t_b[1] + t_b[3] // 2)
                    <= (touch_point[1] - self.y)
                    <= (t_b[1] + t_b[3])
                ):  # in lower half of touch_boundary
                    self.value = self.value - 1

            self._pressed = True  # update the state variable
            self._last_pressed = (
                time.monotonic()
            )  # value changed, so update cool_down timer

    def released(self):
        """Response function when the Control is released. Resets the state variables
        for handling situation when ``cool_down`` is < 0 that requires `released()` before
        reacting another another `selected()`."""
        self._pressed = False

    @property
    def value(self):
        """The value index displayed on the widget. For the setter, the input can
         either be an `int` index into the ``value_list`` or
         can be a `str` that matches one of the items in the ``value_list``.  If `int`,
         the value will be set based on the modulus of the input ``new_value``.

        :return: int
        """

        return self._value

    @value.setter
    def value(self, new_value):  # Set the value based on the index or on the string.
        if isinstance(new_value, str):  # for an input string, search the value_list
            try:
                new_value = self.value_list.index(new_value)
            except ValueError:
                print(
                    'ValueError: Value "{}" not found in value_list.'.format(new_value)
                )
                return None

        new_value = new_value % len(self.value_list)  # Update the value
        if new_value != self._value:
            self._update_value(new_value)
            self._value = new_value
        return self._value


# draw_position - Draws two bitmaps into the target bitmap with offsets.
# Allows values < 0.0 and > 1.0 for "springy" easing functions
def _draw_position(
    target_bitmap,
    bitmap1,
    bitmap1_offset,
    bitmap2,
    bitmap2_offset,
    position=0.0,
    horizontal=True,
):

    x_offset1 = bitmap1_offset[0]
    y_offset1 = bitmap1_offset[1]
    x_offset2 = bitmap2_offset[0]
    y_offset2 = bitmap2_offset[1]

    if position == 0.0:
        target_bitmap.fill(0)
        target_bitmap.blit(x_offset1, y_offset1, bitmap1)
        return
    if position == 1.0:
        target_bitmap.fill(0)
        target_bitmap.blit(x_offset2, y_offset2, bitmap2)
        return

    if horizontal:
        target_bitmap.fill(0)  # clear the target bitmap
        x_index = round(position * target_bitmap.width)  # find the scroll offset
        _blit_constrained(target_bitmap, x_offset1, y_offset1, bitmap1, x1=x_index)
        _blit_constrained(
            target_bitmap,
            target_bitmap.width - x_index + x_offset2,
            y_offset2,
            bitmap2,
            x1=0,
            x2=x_index,
        )

    else:
        target_bitmap.fill(0)
        y_index = round(position * target_bitmap.height)

        _blit_constrained(target_bitmap, x_offset1, y_offset1, bitmap1, y1=y_index)
        _blit_constrained(
            target_bitmap,
            x_offset2,
            target_bitmap.height - y_index + y_offset2,
            bitmap2,
            y1=0,
            y2=y_index,
        )


# pylint: disable=invalid-name

# _blit_constrained: Copies bitmaps with constraints to the dimensions
def _blit_constrained(target, x, y, source, x1=None, y1=None, x2=None, y2=None):
    if x1 is None:
        x1 = 0
    if y1 is None:
        y1 = 0
    if x2 is None:
        x2 = source.width
    if y2 is None:
        y2 = source.height

    if x < 0:
        x1 -= x  # offset the clip region in positive direction
        x2 -= x
        x = 0
    if x1 < 0:
        x = x - x1
        x1 = 0  # move to origin
    if x2 > source.width:
        x2 = source.width

    if y < 0:
        y1 -= y  # offset the clip region
        y2 -= y
        y = 0
    if y1 < 0:
        y = y - y1
        y1 = 0
    if y2 > source.height:
        y2 = source.height

    if (
        (x > target.width)
        or (y > target.height)
        or (x1 > source.width)
        or (y1 > source.height)
    ):
        return

    target.blit(x, y, source, x1=x1, y1=y1, x2=x2, y2=y2)


# _animate_bitmap - performs animation of scrolling between two bitmaps
def _animate_bitmap(
    display,
    target_bitmap,
    bitmap1,
    bitmap1_offset,
    bitmap2,
    bitmap2_offset,
    start_position,
    end_position,
    animation_time,
    horizontal,
):

    start_time = time.monotonic()

    if start_position > end_position:  # direction is decreasing: "out"
        [bitmap2, bitmap1] = [bitmap1, bitmap2]
        [bitmap2_offset, bitmap1_offset] = [bitmap1_offset, bitmap2_offset]
        easing_function = easeout  # use the "out" easing function

    else:  # direction is increasing: "in"
        easing_function = easein  # use the "in" easing function

    display.auto_refresh = False
    _draw_position(
        target_bitmap,
        bitmap1,
        bitmap1_offset,
        bitmap2,
        bitmap2_offset,
        position=start_position,
        horizontal=horizontal,
    )
    display.auto_refresh = True

    while True:

        this_time = time.monotonic()
        target_position = (
            start_position
            + (end_position - start_position)
            * (this_time - start_time)
            / animation_time
        )

        display.auto_refresh = False
        if (this_time - start_time) < animation_time:
            display.auto_refresh = False
            _draw_position(
                target_bitmap,
                bitmap1,
                bitmap1_offset,
                bitmap2,
                bitmap2_offset,
                position=easing_function(target_position),
                horizontal=horizontal,
            )
            display.auto_refresh = True
        else:

            _draw_position(
                target_bitmap,
                bitmap1,
                bitmap1_offset,
                bitmap2,
                bitmap2_offset,
                position=end_position,
                horizontal=horizontal,
            )

            break
        display.auto_refresh = True
    display.auto_refresh = True



# SPDX-FileCopyrightText: 2021 Kevin Matocha
#
# SPDX-License-Identifier: MIT
"""
Creates two animated icons with touch response: zoom and shrink animations.
"""
import time
import board
import displayio
import adafruit_touchscreen
from adafruit_displayio_layout.widgets.icon_animated import IconAnimated

display = board.DISPLAY

ts = adafruit_touchscreen.Touchscreen(
    board.TOUCH_XL,
    board.TOUCH_XR,
    board.TOUCH_YD,
    board.TOUCH_YU,
    calibration=((5200, 59000), (5800, 57000)),
    size=(display.width, display.height),
)


IconAnimated.init_class(
    display, max_scale=1.5, max_icon_size=(48, 48), max_color_depth=255
)

icon_zoom = IconAnimated(
    "Zoom",
    "icons/Play_48x48_small.bmp",
    x=50,
    y=40,
    on_disk=True,
    scale=1.5,  # zoom animation
    angle=5,
)

icon_shrink = IconAnimated(
    "Shrink",
    "icons/Play_48x48_small.bmp",
    x=180,
    y=40,
    on_disk=True,
    scale=0.7,  # shrink animation
    angle=-10,
)

icons = [icon_zoom, icon_shrink]

main_group = displayio.Group()
main_group.append(icon_zoom)
main_group.append(icon_shrink)

display.show(main_group)


COOLDOWN_TIME = 0.25
LAST_PRESS_TIME = -1

display.auto_refresh = True

while True:
    time.sleep(0.05)
    p = ts.touch_point
    if p:
        _now = time.monotonic()
        if _now - LAST_PRESS_TIME > COOLDOWN_TIME:
            for icon in icons:
                if icon.contains(p):
                    icon.zoom_animation(p)
                    LAST_PRESS_TIME = time.monotonic()

    else:
        for icon in icons:
            icon.zoom_out_animation(p)



# SPDX-FileCopyrightText: 2021 Jose David M.
#
# SPDX-License-Identifier: MIT
"""

`cartesian`
================================================================================
A cartesian plane widget for displaying graphical information.

* Author(s): Jose David M.

Implementation Notes
--------------------

**Hardware:**

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

"""

# pylint: disable=too-many-lines, too-many-instance-attributes, too-many-arguments
# pylint: disable=too-many-locals, too-many-statements

import displayio
import terminalio
from adafruit_display_text import bitmap_label
import vectorio
from adafruit_displayio_layout.widgets.widget import Widget
from adafruit_displayio_layout.widgets import rectangle_helper

try:
    import bitmaptools
except ImportError:
    pass

try:
    from typing import Tuple
except ImportError:
    pass


class Cartesian(Widget):
    """A cartesian widget.  The origin is set using ``x`` and ``y``.

    :param int x: x position of the plane origin
    :param int y: y position of the plane origin

    :param int background_color: background color to use defaults to black (0x000000)
    :param int width: requested width, in pixels.
    :param int height: requested height, in pixels.

    :param (int, int) xrange: X axes range. Defaults to (0, 100)
    :param (int, int) yrange: Y axes range. Defaults to (0, 100)

    :param int axes_color: axes lines color defaults to white (0xFFFFFF)
    :param int axes_stroke: axes lines thickness in pixels defaults to 2

    :param int major_tick_stroke: tick lines thickness in pixels defaults to 1
    :param int major_tick_length: tick lines length in pixels defaults to 5

    :param terminalio.FONT tick_label_font: tick label text font
    :param int font_color: font color. Defaults to white (0xFFFFFF)

    :param int pointer_radius: pointer radius in pixels defaults to 1
    :param int pointer_color: pointer color. Defaults to white (0xFFFFFF)

    :param bool subticks: inclusion of subticks in the plot area. Default to False

    :param int nudge_x: movement in pixels in the x direction to move the origin.
     Defaults to 0
    :param int nudge_y: movement in pixels in the y direction to move the origin.
     Defaults to 0


    **Quickstart: Importing and using Cartesian**

    Here is one way of importing the `Cartesian` class so you can use it as
    the name ``Plane``:

    .. code-block:: python

        from adafruit_displayio_layout.widgets.cartesian import Cartesian as Plane

    Now you can create a plane at pixel position x=20, y=30 using:

    .. code-block:: python

        my_plane=Plane(x=20, y=30) # instance the plane at x=20, y=30

    Once you setup your display, you can now add ``my_plane`` to your display using:

    .. code-block:: python

        display.show(my_plane) # add the group to the display

    If you want to have multiple display elements, you can create a group and then
    append the plane and the other elements to the group.  Then, you can add the full
    group to the display as in this example:

    .. code-block:: python

        my_plane= Plane(20, 30) # instance the plane at x=20, y=30
        my_group = displayio.Group() # make a group
        my_group.append(my_plane) # Add my_plane to the group

        #
        # Append other display elements to the group
        #

        display.show(my_group) # add the group to the display


    **Summary: Cartesian Features and input variables**

    The `cartesian` widget has some options for controlling its position, visible appearance,
    and scale through a collection of input variables:

        - **position**: ``x``, ``y``, ``anchor_point``, ``anchored_position`` and
          ``nudge_x``, ``nudge_y``. Nudge parameters are used to account for the float and int
          conversions required to display different ranges and values. Conversion are required
          as displays work in integers and not floats

        - **size**: ``width`` and ``height``

        - **color**: ``axes_color``, ``font_color``, ``tick_color``, ``pointer_color``

        - **background color**: ``background_color``

        - **linewidths**: ``axes_stroke`` and ``major_tick_stroke``

        - **range**: ``xrange`` and ``yrange`` This is the range in absolute units.
          For example, when using (20-90), the X axis will start at 20 finishing at 90.
          However the height of the graph is given by the height parameter. The scale
          is handled internal to provide a 1:1 experience when you update the graph.


    .. figure:: cartesian.gif
       :scale: 100 %
       :figwidth: 50%
       :align: center
       :alt: Diagram of the cartesian widget with the pointer in motion.

       This is a diagram of a cartesian widget with the pointer moving in the
       plot area.

    .. figure:: cartesian_zones.png
       :scale: 100 %
       :figwidth: 50%
       :align: center
       :alt: Diagram of the cartesian widget zones.

       This is a diagram of a cartesian widget showing the different zones.

    .. figure:: cartesian_explanation.png
       :scale: 100 %
       :figwidth: 50%
       :align: center
       :alt: Diagram of the cartesian widget localisation.

       This is a diagram of a cartesian widget showing localisation scheme.

    """

    def __init__(
        self,
        background_color: int = 0x000000,
        xrange: Tuple[int, int] = (0, 100),
        yrange: Tuple[int, int] = (0, 100),
        axes_color: int = 0xFFFFFF,
        axes_stroke: int = 1,
        tick_color: int = 0xFFFFFF,
        major_tick_stroke: int = 1,
        major_tick_length: int = 5,
        tick_label_font=terminalio.FONT,
        font_color: int = 0xFFFFFF,
        pointer_radius: int = 1,
        pointer_color: int = 0xFFFFFF,
        subticks: bool = False,
        nudge_x: int = 0,
        nudge_y: int = 0,
        **kwargs,
    ) -> None:

        super().__init__(**kwargs)

        self._background_color = background_color

        self._axes_line_color = axes_color
        self._axes_line_thickness = axes_stroke

        self._tick_color = tick_color
        if major_tick_stroke not in range(1, 5):
            print("tick thickness must be 1-4 pixels. Defaulting to 1")
            self._tick_line_thickness = 1
        else:
            self._tick_line_thickness = major_tick_stroke

        if major_tick_length not in range(1, 9):
            print("tick length must be 1-10 pixels. Defaulting to 5")
            self._tick_line_height = 5
        else:
            self._tick_line_height = major_tick_length

        self._pointer_radius = pointer_radius
        self._pointer_color = pointer_color

        self._font = tick_label_font
        self._font_color = font_color

        self._font_width = self._get_font_height(self._font, 1)[0]
        self._font_height = self._get_font_height(self._font, 1)[1]

        self._xrange = xrange
        self._normx = (self._xrange[1] - self._xrange[0]) / 100
        self._valuex = self.width / 100
        self._factorx = 100 / (self._xrange[1] - self._xrange[0])

        self._yrange = yrange
        self._normy = (self._yrange[1] - self._yrange[0]) / 100
        self._valuey = self.height / 100
        self._factory = 100 / (self._yrange[1] - self._yrange[0])

        self._tick_bitmap = displayio.Bitmap(
            self._tick_line_thickness, self._tick_line_height, 3
        )
        self._tick_bitmap.fill(1)

        self._subticks = subticks

        axesx_height = (
            2
            + self._axes_line_thickness
            + self._font_height
            + self._tick_line_height // 2
        )

        self._axesx_bitmap = displayio.Bitmap(self.width, axesx_height, 4)
        self._axesx_bitmap.fill(0)

        self._axesy_width = (
            2
            + self._axes_line_thickness
            + self._font_width
            + self._tick_line_height // 2
        )

        self._axesy_bitmap = displayio.Bitmap(self._axesy_width, self.height, 4)
        self._axesy_bitmap.fill(0)

        self._screen_bitmap = displayio.Bitmap(self.width, self.height, 5)
        self._screen_bitmap.fill(5)
        self._screen_palette = displayio.Palette(6)
        self._screen_palette.make_transparent(0)
        self._screen_palette[1] = self._tick_color
        self._screen_palette[2] = self._axes_line_color
        self._screen_palette[3] = 0x990099
        self._screen_palette[4] = 0xFFFFFF
        self._screen_palette[5] = self._background_color

        self._corner_bitmap = displayio.Bitmap(10, 10, 5)
        rectangle_helper(
            0,
            0,
            self._axes_line_thickness,
            self._axes_line_thickness,
            self._corner_bitmap,
            2,
            self._screen_palette,
        )

        self._corner_tilegrid = displayio.TileGrid(
            self._corner_bitmap,
            pixel_shader=self._screen_palette,
            x=-self._axes_line_thickness,
            y=self.height,
        )

        self._axesx_tilegrid = displayio.TileGrid(
            self._axesx_bitmap,
            pixel_shader=self._screen_palette,
            x=0,
            y=self.height,
        )

        self._axesy_tilegrid = displayio.TileGrid(
            self._axesy_bitmap,
            pixel_shader=self._screen_palette,
            x=-self._axesy_width,
            y=0,
        )

        self._screen_tilegrid = displayio.TileGrid(
            self._screen_bitmap,
            pixel_shader=self._screen_palette,
            x=0,
            y=0,
        )

        self._nudge_x = nudge_x
        self._nudge_y = nudge_y

        self._draw_axes()
        self._draw_ticks()

        self.append(self._axesx_tilegrid)
        self.append(self._axesy_tilegrid)
        self.append(self._screen_tilegrid)
        self.append(self._corner_tilegrid)

        self._update_line = True

        self._pointer = None
        self._circle_palette = None
        self._pointer_vector_shape = None
        self.plot_line_point = None

    @staticmethod
    def _get_font_height(font, scale: int) -> Tuple[int, int]:
        if hasattr(font, "get_bounding_box"):
            font_height = int(scale * font.get_bounding_box()[1])
            font_width = int(scale * font.get_bounding_box()[0])
        elif hasattr(font, "ascent"):
            font_height = int(scale * font.ascent + font.ascent)
            font_width = 12
        else:
            font_height = 12
            font_width = 12
        return font_width, font_height

    def _draw_axes(self) -> None:
        # Draw x axes line
        rectangle_helper(
            0,
            0,
            self._axes_line_thickness,
            self.width,
            self._axesx_bitmap,
            2,
            self._screen_palette,
            True,
        )

        # Draw y axes line
        rectangle_helper(
            self._axesy_width - self._axes_line_thickness,
            0,
            self.height,
            self._axes_line_thickness,
            self._axesy_bitmap,
            2,
            self._screen_palette,
            True,
        )

    def _draw_ticks(self) -> None:
        # ticks definition
        ticks = [10, 30, 50, 70, 90]
        subticks = [20, 40, 60, 80, 100]
        # X axes ticks
        for i in range(10, 100, 10):
            text_tick = str(round(self._xrange[0]) + round(i * self._normx))
            text_dist = int(self._valuex * i)
            if i in ticks:
                shift_label_x = len(text_tick) * self._font_width
                tick_text = bitmap_label.Label(
                    self._font,
                    color=self._font_color,
                    text=text_tick,
                    x=text_dist - (shift_label_x // 2),
                    y=self.height
                    + self._axes_line_thickness
                    + self._tick_line_height
                    + self._font_height // 2
                    + 1,
                )
                self.append(tick_text)
                rectangle_helper(
                    text_dist,
                    self._axes_line_thickness,
                    self._tick_line_height,
                    self._tick_line_thickness,
                    self._axesx_bitmap,
                    1,
                    self._screen_palette,
                    True,
                )

            if self._subticks:
                if i in subticks:
                    rectangle_helper(
                        text_dist,
                        self._axes_line_thickness,
                        self._tick_line_height // 2,
                        1,
                        self._axesx_bitmap,
                        1,
                        self._screen_palette,
                        True,
                    )

        # Y axes ticks
        for i in range(10, 100, 10):
            text_tick = str(round(self._yrange[0]) + round(i * self._normy))
            text_dist = int(self._valuey * i)
            if i in ticks:
                shift_label_x = len(text_tick) * self._font_width
                tick_text = bitmap_label.Label(
                    self._font,
                    color=self._font_color,
                    text=text_tick,
                    x=-shift_label_x
                    - self._axes_line_thickness
                    - self._tick_line_height
                    - 2,
                    y=0 + self.height - text_dist,
                )
                self.append(tick_text)
                rectangle_helper(
                    self._axesy_width
                    - self._axes_line_thickness
                    - self._tick_line_height
                    - 1,
                    text_dist,
                    self._tick_line_thickness,
                    self._tick_line_height,
                    self._axesy_bitmap,
                    1,
                    self._screen_palette,
                    True,
                )

            if self._subticks:
                if i in subticks:
                    rectangle_helper(
                        self._axesy_width
                        - self._axes_line_thickness
                        - self._tick_line_height // 2
                        - 1,
                        text_dist,
                        1,
                        self._tick_line_height // 2,
                        self._axesy_bitmap,
                        1,
                        self._screen_palette,
                        True,
                    )

    def _draw_pointers(self, x: int, y: int) -> None:
        self._pointer = vectorio.Circle(self._pointer_radius)
        self._circle_palette = displayio.Palette(2)
        self._circle_palette.make_transparent(0)
        self._circle_palette[1] = self._pointer_color

        self._pointer_vector_shape = vectorio.VectorShape(
            shape=self._pointer,
            pixel_shader=self._circle_palette,
            x=x,
            y=y,
        )
        self.append(self._pointer_vector_shape)

    def update_pointer(self, x: int, y: int) -> None:
        """updater_pointer function
        helper function to update pointer in the plane
        :param int x: ``x`` coordinate in the local plane
        :param int y: ``y`` coordinate in the local plane
        :return: None
        rtype: None
        """
        local_x = int((x - self._xrange[0]) * self._factorx) + self._nudge_x
        local_y = (
            int((self._yrange[0] - y) * self._factory) + self.height + self._nudge_y
        )

        if local_x >= 0 or local_y <= 100:
            if self._update_line:
                self._draw_pointers(local_x, local_y)
                self._update_line = False
            else:
                self._pointer_vector_shape.x = local_x
                self._pointer_vector_shape.y = local_y

    def _set_plotter_line(self) -> None:
        self.plot_line_point = []

    def update_line(self, x: int, y: int) -> None:
        """updater_line function
        helper function to update pointer in the plane
        :param int x: ``x`` coordinate in the local plane
        :param int y: ``y`` coordinate in the local plane
        :return: None
        rtype: None
        """
        local_x = int((x - self._xrange[0]) * self._factorx) + self._nudge_x
        local_y = (
            int((self._yrange[0] - y) * self._factory) + self.height + self._nudge_y
        )
        if x < self._xrange[1] and y < self._yrange[1]:
            if local_x > 0 or local_y < 100:
                if self._update_line:
                    self._set_plotter_line()
                    self.plot_line_point.append((local_x, local_y))
                    self._update_line = False
                else:
                    bitmaptools.draw_line(
                        self._screen_bitmap,
                        self.plot_line_point[-1][0],
                        self.plot_line_point[-1][1],
                        local_x,
                        local_y,
                        1,
                    )



# SPDX-FileCopyrightText: 2020 Kevin Matocha, Tim Cocks
#
# SPDX-License-Identifier: MIT

"""
`grid_layout`
================================================================================

A layout that organizes cells into a grid table structure.


* Author(s): Kevin Matocha, Tim Cocks

Implementation Notes
--------------------

**Hardware:**

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

"""
try:
    # Used only for typing
    from typing import Tuple
except ImportError:
    pass

import math
import displayio

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_DisplayIO_Layout.git"


class GridLayout(displayio.Group):
    """
    A layout that organizes children into a grid table structure.

    :param int x: x location the layout should be placed. Pixel coordinates.
    :param int y: y location the layout should be placed. Pixel coordinates.
    :param int width: Width of the layout in pixels.
    :param int height: Height of the layout in pixels.
    :param tuple grid_size: Size in cells as two ints in a tuple e.g. (2, 2)
    :param int cell_padding: Extra padding space inside each cell. In pixels.
    :param bool divider_lines: Whether or not to draw lines between the cells.
    :param Union[tuple, list] h_divider_line_rows: Row indexes to draw divider
      lines above. Row indexes are 0 based.
    :param Union[tuple, list] v_divider_line_cols: Column indexes to draw divider
      lines before. Column indexes are 0 based.
    :param divider_line_color: The color of the divider lines (in hexadecimal)
    :param tuple cell_anchor_point: Anchor point used within every cell. Needs to
      be a tuple containing two floats between 0.0 and 1.0. Default is (0.0, 0.0)
      which will anchor content to the top left of the cell.

    """

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        x,
        y,
        width,
        height,
        grid_size,
        cell_padding=0,
        divider_lines=False,
        h_divider_line_rows=None,
        v_divider_line_cols=None,
        divider_line_color=0xFFFFFF,
        cell_anchor_point=(0.0, 0.0),
    ):
        super().__init__(x=x, y=y)
        self.x = x
        self.y = y
        self._width = width
        self._height = height
        self.grid_size = grid_size
        self.cell_padding = cell_padding
        self._cell_content_list = []
        self._cell_anchor_point = cell_anchor_point

        self._divider_lines = []
        self._divider_color = divider_line_color
        self.h_divider_line_rows = h_divider_line_rows
        self.v_divider_line_cols = v_divider_line_cols

        self._divider_lines_enabled = (
            (divider_lines is True)
            or (h_divider_line_rows is not None)
            or (v_divider_line_cols is not None)
        )

        if divider_lines:
            if self.h_divider_line_rows is None:
                self.h_divider_line_rows = []
                for _y in range(self.grid_size[1] + 1):
                    self.h_divider_line_rows.append(_y)
            if self.v_divider_line_cols is None:
                self.v_divider_line_cols = []
                for _x in range(self.grid_size[0] + 1):
                    self.v_divider_line_cols.append(_x)
        else:
            if not h_divider_line_rows:
                self.h_divider_line_rows = tuple()
            if not v_divider_line_cols:
                self.v_divider_line_cols = tuple()

        # use at least 1 padding so that content is inside the divider lines
        if cell_padding == 0 and (
            divider_lines or h_divider_line_rows or v_divider_line_cols
        ):
            self.cell_padding = 1

    def _layout_cells(self):
        # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        for cell in self._cell_content_list:
            if cell["content"] not in self:
                grid_size_x = self.grid_size[0]
                grid_size_y = self.grid_size[1]

                grid_position_x = cell["grid_position"][0]
                grid_position_y = cell["grid_position"][1]

                content_cell_size_x = cell["cell_size"][0]
                content_cell_size_y = cell["cell_size"][1]

                _measured_width = (
                    math.ceil(content_cell_size_x * self._width / grid_size_x)
                    - 2 * self.cell_padding
                )

                _measured_height = (
                    math.ceil(content_cell_size_y * self._height / grid_size_y)
                    - 2 * self.cell_padding
                )

                if hasattr(cell["content"], "resize"):
                    # if it has resize function
                    cell["content"].resize(
                        _measured_width,
                        _measured_height,
                    )
                else:
                    try:
                        # try width and height properties.
                        cell["content"].width = _measured_width
                        cell["content"].height = _measured_height
                    except AttributeError:
                        # This element does not allow setting width and height.
                        # No problem, we'll use whatever size it already is.
                        # _measured_width = cell["content"].width
                        # _measured_height = cell["content"].height

                        pass

                if not hasattr(cell["content"], "anchor_point"):

                    cell["content"].x = (
                        int(grid_position_x * self._width / grid_size_x)
                        + self.cell_padding
                        + int(cell["cell_anchor_point"][0] * _measured_width)
                        - int(cell["content"].width * cell["cell_anchor_point"][0])
                    )
                    cell["content"].y = (
                        int(grid_position_y * self._height / grid_size_y)
                        + self.cell_padding
                        + int(cell["cell_anchor_point"][1] * _measured_height)
                        - int(cell["content"].height * cell["cell_anchor_point"][1])
                    )
                else:
                    cell["content"].anchor_point = cell["cell_anchor_point"]
                    cell["content"].anchored_position = (
                        int(grid_position_x * self._width / grid_size_x)
                        + self.cell_padding
                        + (cell["cell_anchor_point"][0] * _measured_width),
                        int(grid_position_y * self._height / grid_size_y)
                        + self.cell_padding
                        + (cell["cell_anchor_point"][1] * _measured_height),
                    )

                self.append(cell["content"])

                if self._divider_lines_enabled:
                    palette = displayio.Palette(2)
                    palette[0] = self._divider_color
                    palette[1] = self._divider_color

                    if not hasattr(cell["content"], "anchor_point"):
                        _bottom_line_loc_y = (
                            (cell["content"].y + _measured_height + self.cell_padding)
                            - 1
                            - int(cell["cell_anchor_point"][1] * _measured_height)
                            + int(cell["content"].height * cell["cell_anchor_point"][1])
                        )

                        _bottom_line_loc_x = (
                            cell["content"].x
                            - self.cell_padding
                            - int(cell["cell_anchor_point"][0] * _measured_width)
                            + int(cell["content"].width * cell["cell_anchor_point"][0])
                        )

                        _top_line_loc_y = (
                            cell["content"].y
                            - self.cell_padding
                            - int(cell["cell_anchor_point"][1] * _measured_height)
                            + int(cell["content"].height * cell["cell_anchor_point"][1])
                        )

                        _top_line_loc_x = (
                            cell["content"].x
                            - self.cell_padding
                            - int(cell["cell_anchor_point"][0] * _measured_width)
                            + int(cell["content"].width * cell["cell_anchor_point"][0])
                        )

                        _right_line_loc_y = (
                            cell["content"].y
                            - self.cell_padding
                            - int(cell["cell_anchor_point"][1] * _measured_height)
                            + int(cell["content"].height * cell["cell_anchor_point"][1])
                        )

                        _right_line_loc_x = (
                            (cell["content"].x + _measured_width + self.cell_padding)
                            - 1
                            - int(cell["cell_anchor_point"][0] * _measured_width)
                            + int(cell["content"].width * cell["cell_anchor_point"][0])
                        )
                    else:
                        _bottom_line_loc_y = (
                            cell["content"].anchored_position[1]
                            + _measured_height
                            + self.cell_padding
                            - (cell["cell_anchor_point"][1] * _measured_height)
                        ) - 1
                        _bottom_line_loc_x = (
                            cell["content"].anchored_position[0]
                            - self.cell_padding
                            - (cell["cell_anchor_point"][0] * _measured_width)
                        )

                        _top_line_loc_y = (
                            cell["content"].anchored_position[1]
                            - self.cell_padding
                            - (cell["cell_anchor_point"][1] * _measured_height)
                        )
                        _top_line_loc_x = (
                            cell["content"].anchored_position[0]
                            - self.cell_padding
                            - (cell["cell_anchor_point"][0] * _measured_width)
                        )

                        _right_line_loc_y = (
                            cell["content"].anchored_position[1]
                            - self.cell_padding
                            - (cell["cell_anchor_point"][1] * _measured_height)
                        )
                        _right_line_loc_x = (
                            (
                                cell["content"].anchored_position[0]
                                + _measured_width
                                + self.cell_padding
                            )
                            - 1
                            - (cell["cell_anchor_point"][0] * _measured_width)
                        )

                    _horizontal_divider_line = displayio.Shape(
                        _measured_width + (2 * self.cell_padding),
                        1,
                        mirror_x=False,
                        mirror_y=False,
                    )

                    _bottom_divider_tilegrid = displayio.TileGrid(
                        _horizontal_divider_line,
                        pixel_shader=palette,
                        y=_bottom_line_loc_y,
                        x=_bottom_line_loc_x,
                    )

                    _top_divider_tilegrid = displayio.TileGrid(
                        _horizontal_divider_line,
                        pixel_shader=palette,
                        y=_top_line_loc_y,
                        x=_top_line_loc_x,
                    )

                    _vertical_divider_line = displayio.Shape(
                        1,
                        _measured_height + (2 * self.cell_padding),
                        mirror_x=False,
                        mirror_y=False,
                    )

                    _left_divider_tilegrid = displayio.TileGrid(
                        _vertical_divider_line,
                        pixel_shader=palette,
                        y=_top_line_loc_y,
                        x=_top_line_loc_x,
                    )

                    _right_divider_tilegrid = displayio.TileGrid(
                        _vertical_divider_line,
                        pixel_shader=palette,
                        y=_right_line_loc_y,
                        x=_right_line_loc_x,
                    )

                    for line_obj in self._divider_lines:
                        self.remove(line_obj["tilegrid"])

                    """
                    Only use bottom divider lines on the bottom row. All
                    other rows rely on top divder lines of the row beneath them.
                    Add the content_cell_size to the grid_position to account for
                    areas larger than 1x1 cells. For 1x1 cells this will equal zero
                    and not change anything.
                    """
                    if (
                        grid_position_y + content_cell_size_y - 1
                    ) == grid_size_y - 1 and (
                        (grid_position_y + content_cell_size_y - 1) + 1
                        in self.h_divider_line_rows
                    ):
                        self._divider_lines.append(
                            {
                                "shape": _horizontal_divider_line,
                                "tilegrid": _bottom_divider_tilegrid,
                            }
                        )

                    """
                    Every cell whose index is in h_divider_line_rows gets
                    a top divider line.
                    """
                    if grid_position_y in self.h_divider_line_rows:
                        self._divider_lines.append(
                            {
                                "shape": _horizontal_divider_line,
                                "tilegrid": _top_divider_tilegrid,
                            }
                        )

                    """
                    Every cell whose index is in v_divider_line_cols gets
                    a left divider line.
                    """
                    if grid_position_x in self.v_divider_line_cols:
                        self._divider_lines.append(
                            {
                                "shape": _horizontal_divider_line,
                                "tilegrid": _left_divider_tilegrid,
                            }
                        )
                    """
                    Only use right divider lines on the right-most column. All
                    other columns rely on left divider lines of the column to their
                    left. Add the content_cell_size to the grid_position to account for
                    areas larger than 1x1 cells. For 1x1 cells this will equal zero
                    and not change anything.
                    """
                    if (
                        grid_position_x + content_cell_size_x - 1
                    ) == grid_size_x - 1 and (
                        (grid_position_x + content_cell_size_x - 1) + 1
                        in self.v_divider_line_cols
                    ):
                        self._divider_lines.append(
                            {
                                "shape": _vertical_divider_line,
                                "tilegrid": _right_divider_tilegrid,
                            }
                        )

                    for line_obj in self._divider_lines:
                        self.append(line_obj["tilegrid"])

    def add_content(
        self, cell_content, grid_position, cell_size, cell_anchor_point=None
    ):
        """Add a child to the grid.

        :param cell_content: the content to add to this cell e.g. label, button, etc...
         Group subclass that have width and height properties can be used.
        :param tuple grid_position: where in the grid it should go. Tuple with
         x,y coordinates in grid cells. e.g. (1,0)
        :param tuple cell_size: the size and shape that the new cell should
         occupy. Width and height in cells inside a tuple e.g. (1, 1)
        :param tuple cell_anchor_point: a tuple of floats between 0.0 and 1.0.
         If passed, this value will override the cell_anchor_point of the GridLayout
         for the single cell having it's content added with this function call. If omitted
         then the cell_anchor_point from the GridLayout will be used.
        :return: None"""

        if cell_anchor_point:
            _this_cell_anchor_point = cell_anchor_point
        else:
            _this_cell_anchor_point = self._cell_anchor_point

        sub_view_obj = {
            "cell_anchor_point": _this_cell_anchor_point,
            "content": cell_content,
            "grid_position": grid_position,
            "cell_size": cell_size,
        }
        self._cell_content_list.append(sub_view_obj)
        self._layout_cells()

    def get_cell(self, cell_coordinates):
        """
        Return a cells content based on the cell_coordinates. Raises
        KeyError if coordinates were not found in the GridLayout.

        :param tuple cell_coordinates: the coordinates to lookup in the grid
        :return: the displayio content object at those coordinates
        """
        for index, cell in enumerate(self._cell_content_list):
            if cell["grid_position"] == cell_coordinates:
                return self._cell_content_list[index]["content"]

        raise KeyError(
            "GridLayout does not contain cell at coordinates {}".format(
                cell_coordinates
            )
        )

    @property
    def cell_size_pixels(self) -> Tuple[int, int]:
        """
        Get the size of a 1x1 cell in pixels. Can be useful for manually
        re-positioning content within cells.

        :return Tuple[int, int]: A tuple containing the (x, y) size in
          pixels of a 1x1 cell in the GridLayout
        """
        return (self._width // self.grid_size[0], self._height // self.grid_size[1])



# SPDX-FileCopyrightText: 2021 Kevin Matocha
#
# SPDX-License-Identifier: MIT
#############################
"""
This is a basic demonstration of a FlipInput widget.
"""

# pylint: disable=invalid-name

import time
import board
import displayio
import adafruit_touchscreen
from adafruit_bitmap_font import bitmap_font
from adafruit_displayio_layout.widgets.flip_input import FlipInput

display = board.DISPLAY  # create the display on the PyPortal,
# otherwise change this to setup the display
# for display chip driver and pinout you have (e.g. ILI9341)

# setup the touchscreen
ts = adafruit_touchscreen.Touchscreen(
    board.TOUCH_XL,
    board.TOUCH_XR,
    board.TOUCH_YD,
    board.TOUCH_YU,
    calibration=((5200, 59000), (5800, 57000)),
    size=(display.width, display.height),
)

# Select the font file for use
font_file = "fonts/DSEG14Classic-Regular-64-ModS.pcf"
my_font = bitmap_font.load_font(font_file)

my_flip1 = FlipInput(
    display,
    anchor_point=[0.0, 0.0],
    anchored_position=[50, 40],
    color=0xFF2222,  # reddish orange color
    value_list=[  # list of month strings, using three characters
        "JAN",
        "FEB",
        "MAR",
        "APR",
        "MAY",
        "JUN",
        "JUL",
        "AUG",
        "SEP",
        "OCT",
        "NOV",
        "DEC",
    ],
    font_scale=5,
    horizontal=False,  # use vertical arrows
    animation_time=0.4,
)

my_flip2 = FlipInput(
    display,
    anchor_point=[0.0, 0.0],
    anchored_position=[220, 40],
    color=0xFF2222,  # reddish orange color
    value_list=["{0:02d}".format(x) for x in range(1, 31 + 1)],
    # use a list of strings from 01 through 31
    # use the {0:02d} format string to always use two digits (e.g. '03')
    font_scale=5,
    horizontal=False,  # use vertical arrows
    animation_time=0.4,
)

my_flip3 = FlipInput(
    display,
    anchor_point=[0.5, 1.0],
    anchored_position=[320 // 2, 240 - 10],
    color=0xFF2222,  # reddish orange color
    value_list=["{}".format(x) for x in range(1985, 2022, 1)],
    # use a list with values of stringsfrom 1985 to 2022
    font=my_font,
    horizontal=True,  # use horizontal arrows
    animation_time=0.8,  # add more time since the animation covers a longer distance
)

# Pick an interesting date to start
#
# You can set the value by direct integer indexes of the list or by a string
# Here are three ways to set the values:
my_flip1.value = 9  # use direct integer indexing to set the value to the 10th month
my_flip2.value = my_flip2.value_list.index("21")  # find the index yourself by
# searching the value_list
my_flip3.value = "2015"  # or set the value based on a string that is in the value_list

# Create the group to display and append the FlipInput widgets
my_group = displayio.Group()
my_group.append(my_flip1)
my_group.append(my_flip2)
my_group.append(my_flip3)

display.show(my_group)  # add high level Group to the display
display.auto_refresh = True

while True:

    p = ts.touch_point
    # print("touch_point p: {}".format(p)) # print the touch point

    if p:  # if touched, check if any of the widgets was triggered
        if my_flip1.contains(p):
            my_flip1.selected(p)
            time.sleep(0.15)  # add a short delay to reduce accidental press
        elif my_flip2.contains(p):
            my_flip2.selected(p)
            time.sleep(0.15)  # add a short delay to reduce accidental press
        elif my_flip3.contains(p):
            my_flip3.selected(p)
            time.sleep(0.15)  # add a short delay to reduce accidental press

    time.sleep(0.01)  # small delay seems to improve touch response
